package common;

import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;


import common.avc.AvcConstants;
import common.mpeg.MpegConstants;
import common.mpeg.MpegUtil;
import common.mpeg.SequenceHelper;
import common.packet.PacketPart;
import common.vc1.Vc1Constants;

public class StreamIdentifier {
	private final HashMap<Byte, Object[]> streamTypeMap=new HashMap<Byte, Object[]>();
	private StreamDescriptor streams[]=new StreamDescriptor[0];
	private StreamAnalyzerI streamAnalyzers[];
	
	public interface StreamAnalyzerI { 
		public StreamDescriptor analyze(MuxBuffer packetSource, int stream, int offset); 
	}
	
	public static class MpegVideoAnalyzer implements StreamAnalyzerI, MpegConstants {
		public static final String CODEC_NAME="MPEG-2"; 
		public static final String FILE_EXTENSION=".mpeg2"; 
			
		public StreamDescriptor analyze(MuxBuffer packetSource, int stream, int offset) {
			//Deal with crappy mux in older versions of TySuiteJ
			while (packetSource.get(offset+2)==B0) {
				offset++;
			}
			
			if (packetSource.get(offset)==B0 && packetSource.get(offset+1)==B0 && packetSource.get(offset+2)==B1) {
				if (packetSource.get(offset+3)==B0) {
					VideoStreamDescriptor descriptor=new VideoStreamDescriptor(stream);
					descriptor.setCodec(CODEC_NAME);
					descriptor.setFileNameExtension(FILE_EXTENSION);
					descriptor.setTsType(MpegConstants.TS_STREAM_TYPE_MPEG2_VIDEO);
					int seqStart=MpegUtil.findStartCode(SEQUENCE_PACKET, packetSource.getByteBuffer(), offset+4);
					if (seqStart > 0) {
						offset=seqStart;
					}
				}
				if (packetSource.get(offset+3)==SEQUENCE_PACKET) {
					SequenceHelper sequenceHelper=new SequenceHelper(packetSource,offset);
					VideoStreamDescriptor descriptor=new VideoStreamDescriptor(stream);
					descriptor.setCodec(CODEC_NAME);
					descriptor.setFileNameExtension(FILE_EXTENSION);
					descriptor.setTsType(MpegConstants.TS_STREAM_TYPE_MPEG2_VIDEO);
					sequenceHelper.buildDescriptor(descriptor);
					return descriptor;
				} else {
					return null;
				}
			}
			return null;
		}
	}
	public static class AdvancedAudioCodecAnalyzer implements StreamAnalyzerI {
		public static final String CODEC_NAME="AAC";
		public static final String FILE_EXTENSION=".aac"; 
		private static int[] FREQ={96000, 88200, 64000, 48000, 44100, 32000, 24000, 22050, 16000, 12000, 11025, 8000, 7350};
		private static String[] CHANNELS={null, "Mono", "Stereo", "3.0", "4.0", "5.0", "5.1", "7.1"};
		
		public StreamDescriptor analyze(MuxBuffer buffer, int streamId, int offset) {
			int value = buffer.getByteBuffer().getShort(offset) & 0xffff;
			if ((value & 0xfff0) == 0xfff0) {
				//ATDS
				byte myBuffer[]=new byte[10];
				buffer.copyTo(offset+1, myBuffer, 0, myBuffer.length);
				BitStream bitStream=new BitStream(myBuffer);
				int sync=bitStream.getBits(4);
				int id=bitStream.getBit();
				int layer=bitStream.getBits(2);
				int protectionAbsent=bitStream.getBit();
				if (layer !=0) {
					return null;
				}
				int profile=bitStream.getBits(2);
				AudioStreamDescriptor asd=getAudioStreamDescriptor(streamId,"ADTS", MpegConstants.TS_STREAM_TYPE_AAC_AUDIO);
				if (!setFreq(asd, bitStream)) {
					return null;
				}
				bitStream.getBit();
				asd.setChannels(CHANNELS[bitStream.getBits(3)]);
				return asd;
			} else if ((value & 0xffe0)==(0x2b7<<5)) {
				//LATM
				byte myBuffer[]=new byte[16];
				buffer.copyTo(offset, myBuffer, 0, myBuffer.length);
				BitStream bitStream=new BitStream(myBuffer);
				int sync=bitStream.getBits(11);
				/*int extensionAudioObjectType =bitStream.getBits(5); //Should be 5
				if (extensionAudioObjectType==5) {
					int sbrPresent=bitStream.getBit();
					if (sbrPresent==1) {
						int extensionSamplingFrequencyIndex=bitStream.getBits(4);
						if (extensionSamplingFrequencyIndex==0xf) {
							int extensionSamplingFrequency=bitStream.getBits(24);
						}
					}
				}*/
				
				int len=bitStream.getBits(13);
				int useSameStreamMux=bitStream.getBit();
				if (useSameStreamMux==1) {
					int version=bitStream.getBit();
					AudioStreamDescriptor asd=getAudioStreamDescriptor(streamId,"LATM", MpegConstants.TS_STREAM_TYPE_LATM_AUDIO);
					if (version==1) {
						int versionA=bitStream.getBit();
						if (versionA==1) {
							//TODO: Figure out how to parse Version A
							return asd;
						}
						int taraBufferFullness=getLatmValue(bitStream);
					}
					int allStreamsSameTimeFraming=bitStream.getBit();
					int numSubFrames=bitStream.getBits(6) + 1;
					int numPrograms=bitStream.getBits(4) + 1;
					int numLayers=bitStream.getBits(3) + 1;
					
					if (numPrograms > 1 || allStreamsSameTimeFraming == 0 || numSubFrames > 1 || numLayers > 1) {
						return null;
					}
					int ascLen=0;
					if (version==1) {
						ascLen=getLatmValue(bitStream);
					}
					int objectType=bitStream.getBits(5);
					if (!setFreq(asd, bitStream)) {
						return null;
					}
				}
			}
			return null;
		}
		
		private AudioStreamDescriptor getAudioStreamDescriptor(int streamId, String postFix, byte tsType) {
			AudioStreamDescriptor asd=new AudioStreamDescriptor(streamId);
			asd.setCodec(CODEC_NAME+"-"+postFix);
			asd.setFileNameExtension(FILE_EXTENSION);
			asd.setTsType(tsType);
			return asd;
		}
		
		private boolean setFreq(AudioStreamDescriptor asd, BitStream bitStream) {				
				int samplingFrequencyIndex=bitStream.getBits(4);
				if (samplingFrequencyIndex>=13) {
					return false;
				}
				asd.setFrequency(FREQ[samplingFrequencyIndex]);
				return true;
		}
		
		/*private boolean setChannels(AudioStreamDescriptor asd, BitStream bitStream) {
			int channelsConfiguration=bitStream.getBits(4);
			asd.setChannels(CHANNELS[channelsConfiguration]);
			return true;
		}*/
		private int getLatmValue(BitStream bitStream) {
			int bytes=bitStream.getBits(2);
			int value=0;
			for (int i=0;i<bytes;i++) {
				value = (value << 8) + bitStream.getBits(8);
			}
			return value;
		}
	}
	
	public static class VC1VideoAnalyzer implements StreamAnalyzerI, Vc1Constants {
		public static final String CODEC_NAME="VC-1"; 
		public static final String FILE_EXTENSION=".vc1"; 

		public StreamDescriptor analyze(MuxBuffer buffer, int stream, int offset) {
			//Check for VC-1 Sequence
			if (buffer.get(offset)==B0 && buffer.get(offset+1)==B0 && buffer.get(offset+2)==B1 && buffer.get(offset+3)==SEQUENCE_START) {
				byte myBuffer[]=new byte[buffer.limit()-offset-4];
				buffer.copyTo(offset+4, myBuffer, 0, myBuffer.length);
				BitStream bitStream=new BitStream(myBuffer);
				VideoStreamDescriptor descriptor=new VideoStreamDescriptor(stream);
				int profile = bitStream.getBits(2);
				descriptor.setCodec(CODEC_NAME);
				descriptor.setProfileName(PROFILE_NAMES[profile]);
				if (profile==3 ) {
					//Advanced Profile
					int level = bitStream.getBits(3);
					descriptor.setLevelName("L"+level);
					int chromaFormat=bitStream.getBits(2);
					int crap=bitStream.getBits(8);
					int postProcessingFlag=bitStream.getBits(1);
					int width=(bitStream.getBits(12) + 1)*2;
					int height=(bitStream.getBits(12) + 1)*2;
					int pulldownFlag=bitStream.getBits(1);
					descriptor.setInterlaced(bitStream.getBits(1));
					int frameCounterFlag=bitStream.getBits(1);
					//Common
					int frameInterpolationFlag=bitStream.getBits(1);
					//Advanced
					int reserved=bitStream.getBits(1);
					int psfFlag=bitStream.getBits(1);
					//System.out.println("Pulldown: "+pulldownFlag + " Interlace: " + interlacedFlag + " FrameCounters: " + frameCounterFlag + " PsF: " + psfFlag);
					int displayExtensionFlag=bitStream.getBits(1);
					if (displayExtensionFlag==1) {
						int displayWidth=bitStream.getBits(14)+1;
						int displayHeight=bitStream.getBits(14)+1;
						int aspectRatioFlag=bitStream.getBits(1);
						if (aspectRatioFlag==1) {
							int aspectRatioId=bitStream.getBits(4);
							int aspectRatioWidth, aspectRatioHeight;
							if (aspectRatioId==15) {
								aspectRatioWidth=bitStream.getBits(8);
								aspectRatioHeight=bitStream.getBits(8);
							} else {
								aspectRatioWidth=ASPECT_RATIO[aspectRatioId][0];
								aspectRatioHeight=ASPECT_RATIO[aspectRatioId][1];
							}
							//This doesn't make sense in VC-1.  If the SAR is present (and not 1:1, then the width != displayWidth);
							//Removing it as not to confuse the UI
							if (width!=displayWidth) {
								descriptor.setPixelAspectRatio(new int[] {aspectRatioWidth,aspectRatioHeight});
							}
						}
						int frameRateFlag=bitStream.getBits(1);
						if (frameRateFlag==1) {
							int explicitFrameRateFlag=bitStream.getBits(1);
							int frameRateNumerator, frameRateDenominator;
							
							if (explicitFrameRateFlag==1) {
								frameRateNumerator=bitStream.getBits(16);
								frameRateDenominator=32;
							} else {
								int frameRateNumeratorId=bitStream.getBits(8);
								frameRateNumerator=FRAME_RATE_NUMERATORS[frameRateNumeratorId];
								int frameRateDenominatorId=bitStream.getBits(4);
								frameRateDenominator=FRAME_RATE_DENOMINATORS[frameRateDenominatorId];
							}
							descriptor.setFrameRateRatio(new int[]{frameRateNumerator,frameRateDenominator});
						}
					}
					descriptor.setWidth(width);
					descriptor.setHeight(height);
				}
				descriptor.setFileNameExtension(FILE_EXTENSION);
				descriptor.setTsType(MpegConstants.TS_STREAM_TYPE_VC1_VIDEO);
				return descriptor;
			}
			return null;
		}
	}
	/*public static class H264VideoAnalyzer implements StreamAnalyzerI, MpegConstants {
		public static final String CODEC_NAME="AVC";
		public StreamDescriptor analyze(MuxBuffer packetSource, int stream, int offset) {
			for (int i=0;i<3;i++) {
				if (packetSource.get(offset+i)!=B0) {
					return null;
				}
			}
			if (packetSource.get(offset+3) != B1) {
				return null;
			}
			if (packetSource.get(offset+4) == (byte)0x9) {
				VideoStreamDescriptor descriptor=new VideoStreamDescriptor(stream);
				descriptor.setCodec(CODEC_NAME);
				descriptor.setFileNameExtension(".m4v");
				descriptor.setTsType(MpegConstants.TS_STREAM_TYPE_AVC_VIDEO);
				return descriptor;
			} else {
				return null;
			}
		}
		
		
	}*/
	public static class LpcmAnalyzer implements StreamAnalyzerI {
		public static final String CODEC_NAME="LPCM";
		private static final byte SYNC_BYTE0=0x10;
		private static final byte SYNC_BYTE1=(byte)0xE0;
		private int LPCM_CHANNELS[]={0,1,0,2,3,4,0,0,0,6};
		public StreamDescriptor analyze(MuxBuffer packetSource, int stream, int offset) {
			//This is weak
			if (packetSource.get(offset+0)==SYNC_BYTE0 && packetSource.get(offset+1)==SYNC_BYTE1) {
				AudioStreamDescriptor descriptor=new AudioStreamDescriptor(stream);
				descriptor.setCodec(CODEC_NAME);
				descriptor.setFileNameExtension(".pcm");
				descriptor.setTsType(MpegConstants.TS_STREAM_TYPE_PCM_AUDIO);
				return descriptor;
			}
			return null;
		}
		
	}
	public static class DolbyDigitalAnalzyer implements StreamAnalyzerI {
		private static final byte SYNC_BYTE0=0x0B;
		private static final byte SYNC_BYTE1=0x77;
		public static final String EAC3_CODEC_NAME="E-AC3";
		public static final String AC3_CODEC_NAME="AC3";
		public static final String AC3_FILE_EXTENSION=".ac3";
		public static final String EAC3_FILE_EXTENSION=".eac3";
		public static final byte[] SYNC={SYNC_BYTE0, SYNC_BYTE1};
		private final int FREQUENCIES[]={48000,44100,32000};
		private final char CHANNELS[]={'2','1','2','3','3','4','4','5'};
		private final int BIT_RATE[]={32,40,48,56,64,80,96,112,128,160,192,224,256,320,384,448,512,576,640};
		private final int NUMBER_OF_BLOCKS[]={1, 2, 3, 6};
		
		private BitStream getBitStream(MuxBuffer packetSource, int offset, int len) {
			byte buffer[]=new byte[len];
			for (int i=0;i<buffer.length;i++) {
				buffer[i]=packetSource.get(offset+i);
			}
			return new BitStream(buffer);
		}
		public StreamDescriptor analyze(MuxBuffer packetSource, int stream, int offset) {
			if (packetSource.get(offset+0)==SYNC_BYTE0 && packetSource.get(offset+1)==SYNC_BYTE1) {
				AudioStreamDescriptor descriptor=new AudioStreamDescriptor(stream);
				//Common for AC3/E-AC3
				int fscod=(packetSource.get(offset+4)&0xC0) >>> 6;
				descriptor.setFrequency(FREQUENCIES[fscod]);
				int bsid=(packetSource.get(offset+5)&0xF8) >>> 3;
				
				int lfeon;
				//int channelFlags;
				int acmod;
				int bitRate;
				descriptor.setFileNameExtension(AC3_FILE_EXTENSION);
				int numberOfBlocks;
				BitStream bitStream;
				if (bsid==4 || bsid==8 || bsid==6) {
					bitStream=getBitStream(packetSource,offset+4,4);
					fscod=bitStream.getBits(2);
					int frmsizecod=bitStream.getBits(6);
					bsid=bitStream.getBits(5);
					int bsmod=bitStream.getBits(3);
					acmod=bitStream.getBits(3);
					
					descriptor.setTsType(MpegConstants.TS_STREAM_TYPE_AC3_AUDIO);
					descriptor.setCodec(AC3_CODEC_NAME);
					//channelFlags=packetSource.get(offset+6)&0xE0;
					//channelIndex=channelFlags >>> 5;
					//System.out.println((lags&0x3F));
					//Only accurate for 48kHz
					bitRate=BIT_RATE[frmsizecod>>>1];
					
					int cmixlev=-1;
					if ((acmod & 0x1)==0x1 && acmod != 0x1) {
						cmixlev=bitStream.getBits(2);
					}
					int surmixlev=-1;
					if ((acmod & 0x4)==0x4) {
						surmixlev=bitStream.getBits(2);
					}
					int dsurmod=-1;
					if (acmod==0x2) {
						dsurmod=bitStream.getBits(2);
					}
					lfeon=bitStream.getBit();
					numberOfBlocks=6;
				} else if (bsid==16){
					//TODO: Remove Test code
					descriptor.setTsType(MpegConstants.TS_STREAM_TYPE_EAC3_AUDIO);
					//descriptor.setTsType((byte)0x89);
					descriptor.setCodec(EAC3_CODEC_NAME);
					bitStream=getBitStream(packetSource, offset+2,6);
					int strmtype=bitStream.getBits(2);
					int substreamid=bitStream.getBits(3);
					int frmsize=bitStream.getBits(11) + 1;
					fscod=bitStream.getBits(2);//Redundant
					int numblkscod;
					if (fscod==0x3) {
						int fscod2=bitStream.getBits(3);
						numblkscod=0x3;
					} else {
						numblkscod=bitStream.getBits(2);
					}
					numberOfBlocks=NUMBER_OF_BLOCKS[numblkscod];
					bitRate=frmsize*descriptor.getFrequency()*16/(numberOfBlocks*256)/1000;
					acmod=bitStream.getBits(3);
					lfeon=bitStream.getBit();
					descriptor.setFileNameExtension(EAC3_FILE_EXTENSION);
					//TODO: Figure out tsStreamType
				} else {
					System.out.println("Uknown AC3 type: " + bsid);
					return null;
				}
				descriptor.setBitRate(bitRate);
				descriptor.setFrameTime(numberOfBlocks*256*((int)MpegConstants.PTS_FREQ)/descriptor.getFrequency());
				descriptor.setChannels(CHANNELS[acmod] + "." + (lfeon==0?'0':'1'));
				return descriptor;
			}
			return null;
		}
	}
	public static class DtsAnalyzer implements StreamAnalyzerI {
		public static final String CODEC_NAME="dts";
		public static final String FILE_EXTENSION=".dts";
		private static final byte SYNC_BYTE0=0x7F;
		private static final byte SYNC_BYTE1=(byte)0xFE;
		private static final byte SYNC_BYTE2=(byte)0x80;
		private static final byte SYNC_BYTE3=0x01;
		public static final byte[] SYNC={SYNC_BYTE0, SYNC_BYTE1, SYNC_BYTE2, SYNC_BYTE3};

		private final int DTS_SAMPLE_RATES[]={0, 8000, 16000, 32000, 0, 0, 11025, 22050, 44100, 0, 0, 12000, 24000, 48000, 96000, 192000};
		private final int DTS_BIT_RATES[]={
			    32, 56, 64, 96, 112, 128, 192, 224, 256, 320, 384,
			    448, 512, 576, 640, 768, 896, 1024, 1152, 1280, 1344,
			    1408, 1411, 1472, 1536, 192,
			    2048, 3072, 3840, -1/*open*/, -2/*variable*/, -3/*lossless*/
			};
		private final int DTS_CHANNELS[]={1,2,2,2,2,3,3,4,4,5,6,7};
		public StreamDescriptor analyze(MuxBuffer packetSource, int stream, int offset) {
			if (packetSource.get(offset)==SYNC_BYTE0 && packetSource.get(offset+1)==SYNC_BYTE1 && packetSource.get(offset+2)==SYNC_BYTE2 && packetSource.get(offset+3)==SYNC_BYTE3) {
				int frameLength=packetSource.get(offset+4) << 6;
				frameLength += (((packetSource.get(offset+5) & 0xFC) >>> 2)+1)*32;
				int frameSize=(packetSource.get(offset+5) & 0x3) << 12;
				frameSize += (packetSource.get(offset+6)&0xFF) << 4;
				frameSize += (((packetSource.get(offset+7)& 0xF0)>>>6) +1) * 8 / 14 * 2;
				int flags=(packetSource.get(offset+7)& 0x0F) <<2;
				flags +=(packetSource.get(offset+8) &0xC0) >>> 6;
				int channels=DTS_CHANNELS[flags];
				int sampleRateType = (packetSource.get(offset+8) & 0x36) >>> 2;
				int bitRateType = (packetSource.get(offset+8) & 0x03) << 3;
				bitRateType += (packetSource.get(offset+9)&0xE0)>>>5;
				int lfeFlag = (packetSource.get(offset+10) & 0x6) >>> 2;
				AudioStreamDescriptor descriptor=new AudioStreamDescriptor(stream);
				descriptor.setCodec(CODEC_NAME);
				descriptor.setBitRate(DTS_BIT_RATES[bitRateType]);
				descriptor.setFrequency(DTS_SAMPLE_RATES[sampleRateType]);
				descriptor.setChannels(channels + "." + lfeFlag);
				descriptor.setFileNameExtension(FILE_EXTENSION);
				//TODO: Figure out TS stream type
				descriptor.setTsType(MpegConstants.TS_STREAM_TYPE_DTS_AUDIO);
				return descriptor;
			}
			return null;
		}
	}
	public static class DtsMasterAudioAnalyzer {
		public static final String CODEC_NAME="dts Master Audio";
		public static final String FILE_EXTENSION=".dma";
		
	}
	public static class MpegAudioAnalyzer implements StreamAnalyzerI {
		public static final String CODEC_NAME="MPEG Audio"; //Need to startsWith()
		private static final byte SYNC_BYTE0=(byte)0xFF;
		private static final int SYNC_BYTE1MASK=0xF0;//Sync + 1 bit for mpeg indicator
		private final int FREQUENCIES[]={44100,48000,32000};
		private final int MP2_BIT_RATE_MULTIPLIER[]={0,32,16,8};
		private final int LAYER_III=1;
		private final int LAYER_II=2;
		private final int VERSION_1=3;
		private final int VERSION_2=2;
		private final int MP1_BIT_RATE[][]=
			{	{0,0,0},
				{32, 32, 32},
				{64, 48, 40},
				{96, 56, 48},
				{128, 64, 56},
				{160, 80, 64},
				{192, 96, 80},
				{224, 112, 96},
				{256, 128, 112},
				{288, 160, 128},
				{320, 192, 160},
				{352, 224, 192},
				{384, 256, 224},
				{416, 320, 256},
				{448, 384, 320}
			};
		
		private int[] syncOffsets={0,4};
		
		public void setSyncOffsets(int[] syncOffsets) {
			this.syncOffsets=syncOffsets;
		}
		private final String CHANNELS[]={"Stereo","Joint Sterio","Dual Channel","Mono"};
		public StreamDescriptor analyze(MuxBuffer packetSource, int stream, int offset) {
			for (int i=0; i<syncOffsets.length; i++) {
				if (packetSource.get(offset + syncOffsets[i])==SYNC_BYTE0 && (packetSource.get(offset+syncOffsets[i]+1)&SYNC_BYTE1MASK)==SYNC_BYTE1MASK) {
					offset+=syncOffsets[i];
					//int id=(packetSource.getByte(1)&0x08)>>>3;
					int version=(packetSource.get(offset+1)&0x18)>>>3; //3=Version 1, 2=Version 2
					if (version==1) {
						return null;
					}
					int layer=(packetSource.get(offset+1)&0x06)>>>1; //3=Layer 1, 2=Layer 2,1=Layer 3
					if (layer==0) {
						return null;
					}
					int bitRateIndex=(packetSource.get(offset+2)&0xF0) >>>4;
					if (bitRateIndex==0xf) {
						return null;
					}
					int freqIndex=(packetSource.get(offset+2)&0x0C)>>>2;
					if (freqIndex==3) {
						return null;
					}
					int modeIndex=(packetSource.get(offset+3)&0xC0)>>>6;
					AudioStreamDescriptor descriptor=new AudioStreamDescriptor(stream);
					descriptor.setCodec(CODEC_NAME + " V" + (4-version) + "L" + (4-layer));
					if (version==VERSION_2) {
						descriptor.setBitRate(32+bitRateIndex*MP2_BIT_RATE_MULTIPLIER[layer]);
					} else if (version==VERSION_1) {
						descriptor.setBitRate(MP1_BIT_RATE[bitRateIndex][3-layer]);
					}
					descriptor.setFrequency(FREQUENCIES[freqIndex]);
					descriptor.setChannels(CHANNELS[modeIndex]);
					String ext=".mpa";
					if (layer==LAYER_II) {
						ext=".mp2";
					} else if (layer==LAYER_III) {
						ext=".mp3";
					}
					descriptor.setFileNameExtension(ext);
					
					//Technically, this is wrong for MPEG1, but WMP won't play MPEG1 in TS Stream.
					//Assuming MPEG2 is backwards compatible.
					descriptor.setTsType(MpegConstants.TS_STREAM_TYPE_MPEG2_AUDIO);
					return descriptor;
				}
			}
			return null;
		}
		
	}
	public static class DolbyTrueHdAnalyzer implements StreamAnalyzerI {
		public static final String CODEC_NAME="TrueHD";
		public static final String FILE_EXTENSION=".thd";
		private static final byte B0=0;
		public StreamDescriptor analyze(MuxBuffer buffer, int streamId, int offset) {
			//This is really week
			if (streamId >=0xB0 && streamId < 0xC0) {
				AudioStreamDescriptor descriptor=new AudioStreamDescriptor(streamId);
				descriptor.setCodec(CODEC_NAME);
				descriptor.setFileNameExtension(FILE_EXTENSION);
				descriptor.setTsType(MpegConstants.TS_STREAM_TYPE_TRUEHD_AUDIO);
				return descriptor;
			}
			return null;
		}
		
	}
	public StreamIdentifier() {
		this(new StreamAnalyzerI[] {new MpegVideoAnalyzer(), new VC1VideoAnalyzer(), new AvcVideoAnalyzer(), new DolbyDigitalAnalzyer(), new DtsAnalyzer(), new LpcmAnalyzer(), new DolbyTrueHdAnalyzer(), new MpegAudioAnalyzer(), new AdvancedAudioCodecAnalyzer()});
	}
	
	public StreamIdentifier(StreamAnalyzerI[] streamAnalyzers) {
		this.streamAnalyzers=streamAnalyzers;
		
		//BluRay stream types
		//Format (Key: StreamType, Value:{CodecName, Codec ClassificationId, DefaultExtension, SubStreamInfo{subStreamId,subStreamType}} 
		streamTypeMap.put(Byte.valueOf(MpegConstants.TS_STREAM_TYPE_MPEG1_VIDEO), new Object[]{"MPEG-1 Video",Integer.valueOf(StreamDescriptor.CLASSIFICATION_VIDEO),".mpv"});
		streamTypeMap.put(Byte.valueOf(MpegConstants.TS_STREAM_TYPE_MPEG2_VIDEO), new Object[]{"MPEG-2 Video",Integer.valueOf(StreamDescriptor.CLASSIFICATION_VIDEO),".mpv"});
		streamTypeMap.put(Byte.valueOf(MpegConstants.TS_STREAM_TYPE_MPEG1_AUDIO), new Object[]{"MPEG-1 Layer II Audio",Integer.valueOf(StreamDescriptor.CLASSIFICATION_AUDIO),".mp2"});
		streamTypeMap.put(Byte.valueOf(MpegConstants.TS_STREAM_TYPE_MPEG2_AUDIO), new Object[]{"MPEG-2 Layer II Audio",Integer.valueOf(StreamDescriptor.CLASSIFICATION_AUDIO),".mp2"});
		streamTypeMap.put(Byte.valueOf(MpegConstants.TS_STREAM_TYPE_AAC_AUDIO), new Object[]{"Advanced Audio Codec",Integer.valueOf(StreamDescriptor.CLASSIFICATION_AUDIO),AdvancedAudioCodecAnalyzer.FILE_EXTENSION});
		streamTypeMap.put(Byte.valueOf(MpegConstants.TS_STREAM_TYPE_LATM_AUDIO), new Object[]{"AAC-LATM",Integer.valueOf(StreamDescriptor.CLASSIFICATION_AUDIO),AdvancedAudioCodecAnalyzer.FILE_EXTENSION});
		streamTypeMap.put(Byte.valueOf(MpegConstants.TS_STREAM_TYPE_AVC_VIDEO), new Object[]{"AVC Video",Integer.valueOf(StreamDescriptor.CLASSIFICATION_VIDEO),AvcConstants.FILE_EXTENSION});
		streamTypeMap.put(Byte.valueOf(MpegConstants.TS_STREAM_TYPE_PCM_AUDIO), new Object[]{"PCM Audio",Integer.valueOf(StreamDescriptor.CLASSIFICATION_AUDIO),".pcm"});
		streamTypeMap.put(Byte.valueOf(MpegConstants.TS_STREAM_TYPE_EAC3_AUDIO), new Object[]{"Dolby Digital Plus Audio",Integer.valueOf(StreamDescriptor.CLASSIFICATION_AUDIO),DolbyDigitalAnalzyer.EAC3_FILE_EXTENSION, new int[]{0x71, MpegConstants.TS_STREAM_TYPE_AC3_AUDIO}});
		streamTypeMap.put(Byte.valueOf(MpegConstants.TS_STREAM_TYPE_AC3_AUDIO), new Object[]{"Dolby Digital Audio",Integer.valueOf(StreamDescriptor.CLASSIFICATION_AUDIO),DolbyDigitalAnalzyer.AC3_FILE_EXTENSION});
		streamTypeMap.put(Byte.valueOf(MpegConstants.TS_STREAM_TYPE_DTS_AUDIO), new Object[]{"DTS Audio",Integer.valueOf(StreamDescriptor.CLASSIFICATION_AUDIO),DtsAnalyzer.FILE_EXTENSION});
		streamTypeMap.put(Byte.valueOf(MpegConstants.TS_STREAM_TYPE_TRUEHD_AUDIO), new Object[]{"Dolby TrueHD Audio",Integer.valueOf(StreamDescriptor.CLASSIFICATION_AUDIO),DolbyTrueHdAnalyzer.FILE_EXTENSION, new int[]{0x76, MpegConstants.TS_STREAM_TYPE_AC3_AUDIO}});
		streamTypeMap.put(Byte.valueOf(MpegConstants.TS_STREAM_TYPE_DTS_HIRES_AUDIO), new Object[]{"DTS-High Res. Audio",Integer.valueOf(StreamDescriptor.CLASSIFICATION_AUDIO),".dhr", new int[]{0x71, MpegConstants.TS_STREAM_TYPE_DTS_AUDIO}});
		streamTypeMap.put(Byte.valueOf(MpegConstants.TS_STREAM_TYPE_DTSHD_AUDIO), new Object[]{"DTS-HD Master Audio",Integer.valueOf(StreamDescriptor.CLASSIFICATION_AUDIO),DtsMasterAudioAnalyzer.FILE_EXTENSION, new int[]{0x71, MpegConstants.TS_STREAM_TYPE_DTS_AUDIO}});
		streamTypeMap.put(Byte.valueOf(MpegConstants.TS_STREAM_TYPE_DTS_EXPRESS_AUDIO), new Object[]{"DTS Express Audio",Integer.valueOf(StreamDescriptor.CLASSIFICATION_AUDIO),DtsAnalyzer.FILE_EXTENSION});
		streamTypeMap.put(Byte.valueOf(MpegConstants.TS_STREAM_TYPE_SUBTITLE), new Object[]{"Subtitle Video",Integer.valueOf(StreamDescriptor.CLASSIFICATION_SUBTITLE),".sub"});
		streamTypeMap.put(Byte.valueOf(MpegConstants.TS_STREAM_TYPE_INTERACTIVE_STREAM), new Object[]{"Interactive Stream",Integer.valueOf(StreamDescriptor.CLASSIFICATION_UNKNOWN),".raw"});
		streamTypeMap.put(Byte.valueOf(MpegConstants.TS_STREAM_TYPE_VC1_VIDEO), new Object[]{"VC-1 Video",Integer.valueOf(StreamDescriptor.CLASSIFICATION_VIDEO),VC1VideoAnalyzer.FILE_EXTENSION});
	}
	
	private StreamDescriptor addStream(StreamDescriptor streamDescriptor) {
		for (int i=0;i<streams.length;i++) {
			if (streams[i].getStreamId()==streamDescriptor.getStreamId()) {
				StreamDescriptor old=streams[i];
				streams[i]=streamDescriptor;
				return old;
			}
		}
		StreamDescriptor newStreams[]=new StreamDescriptor[streams.length + 1];
		System.arraycopy(streams, 0, newStreams, 0, streams.length);
		newStreams[streams.length]=streamDescriptor;
		streams=newStreams;
		return null;
	}
	public StreamDescriptor getStreamDescriptor(int stream, PacketPart packetPart, int offset) {
		//Skip the PES header
		StreamDescriptor streamDescriptor=null;
		for (int a=0; a<streamAnalyzers.length;a++) {
			streamDescriptor=streamAnalyzers[a].analyze(packetPart, stream, offset);
			if (streamDescriptor != null) {
				addStream(streamDescriptor);
				//System.out.print(Util.longToHex(stream,2) + ": ") ;
				return streamDescriptor;
			}
		}
		return streamDescriptor;
	}
	public boolean identified(int stream) {
		return getStreamDescriptor(stream) !=null;
	}
	public StreamDescriptor getStreamDescriptor(int stream) {
		for (int i=0;i<streams.length;i++) {
			if (streams[i].getStreamId()==stream) {
				return streams[i];
			}
		}
		return null;
	}
	public StreamDescriptor identifyStream(int stream, PacketPart packetPart) throws IOException {
		StreamDescriptor streamDescriptor=getStreamDescriptor(stream);
		//for (int i=0;i<packetParts.size();i++) {
		if (streamDescriptor==null || streamDescriptor.getClass()==StreamDescriptor.class) {
			streamDescriptor=getStreamDescriptor(stream, packetPart, 0);
			if (streamDescriptor==null) {
				/*System.out.print(Util.longToHex(stream,2) + ": ") ;
				for (int j=0;j<packet.getLen() && j<32;j++) {
					Util.printHex(packet.getByte(j));
				}
				System.out.println();*/
				streamDescriptor=new StreamDescriptor(stream);
			} else {
				addStream(streamDescriptor);
			}
		}
		return streamDescriptor;
	}
	
	public static void main(String args[]) {
	}
	public void streamEnded() throws IOException {}
	
	public VideoStreamDescriptor getVideo() {
		for (int i=0;i<streams.length;i++) {
			if (streams[i].getClassification()==StreamDescriptor.CLASSIFICATION_VIDEO) {
				return (VideoStreamDescriptor)streams[i];
			}
		}
		return null;
	}
	public AudioStreamDescriptor[] getAudio() {
		ArrayList<AudioStreamDescriptor> audioList=new ArrayList<AudioStreamDescriptor>(streams.length);
		for (int i=0;i<streams.length;i++) {
			if (streams[i].getClassification()==StreamDescriptor.CLASSIFICATION_AUDIO) {
				audioList.add((AudioStreamDescriptor)streams[i]);
			}
		}
		AudioStreamDescriptor audioDescriptors[]=new AudioStreamDescriptor[audioList.size()];
		return (AudioStreamDescriptor[])audioList.toArray(audioDescriptors);
	}
	public StreamDescriptor getTsStreamDescriptor(int pid, byte type) {
		Object data[]=(Object[])streamTypeMap.get(Byte.valueOf(type));
		StreamDescriptor streamDescriptor;
		if (data !=null) {
			switch (((Integer)data[1]).intValue()) {
				case StreamDescriptor.CLASSIFICATION_VIDEO:
					streamDescriptor=new VideoStreamDescriptor(pid);
					break;
				case StreamDescriptor.CLASSIFICATION_AUDIO:
					streamDescriptor=new AudioStreamDescriptor(pid);
					break;
				default:
					streamDescriptor=new StreamDescriptor(pid, ((Integer)data[1]).intValue());
			}
			streamDescriptor.setTsType(type);
			streamDescriptor.setCodec((String)data[0]);
			streamDescriptor.setFileNameExtension((String)data[2]);
			if (data.length==4) {
				int subStreamData[]=(int[])data[3];
				streamDescriptor.setCoreStreamDescriptor(getTsStreamDescriptor(subStreamData[0],(byte)subStreamData[1]));
			}
		} else {
			streamDescriptor=new StreamDescriptor(pid, StreamDescriptor.CLASSIFICATION_UNKNOWN);
			streamDescriptor.setCodec("Uknown type: 0x" + Integer.toString(type&0xff, 16));
		}
		return streamDescriptor;
	}
}
