package common.mpeg;

import java.io.File;
import java.io.IOException;
import java.io.RandomAccessFile;
import java.nio.ByteBuffer;

import common.MuxBuffer;
import common.Util;

public class MpegVideoPacket {
	public static final int MINIMUM_VIDEO_PACKET_SIZE=48;
	public static final int ASPECT_RATIO_FULL=2;
	public static final int ASPECT_RATIO_WIDE=3;
	
	private static final String ASPECT_RATIO_TEXT[]={"", " 1x1", " 4x3", " 16x9", " 2.21x1"};
	
	public static final byte SEQUENCE_PACKET=(byte)0xB3;
	private static final byte BYTE0=0;
	private static final byte BYTE1=1;
	private static final byte STUFFING_BYTE=(byte)0xFF;
	
	private static final byte DTS_PRESENT=(byte)0x40;
	
	private SequenceHelper sequenceHelper=new SequenceHelper();
	
	private long pts;
	private long dts;
	private int length;
	
	private int headerLength;
	
	private int sequencePacketStart;
	
	public static boolean containsSequenceHeader(ByteBuffer buffer, int offset, int size) {
		return findSequence(buffer, offset, size) >= 0;
	}
	private static int findSequence(ByteBuffer buffer, int offset, int size) {
		if (size > MINIMUM_VIDEO_PACKET_SIZE) {
			int headerLength=buffer.get(offset+8) & 0xFF;
			int firstSubPacketPos = offset + 9 + headerLength;
			
			while (buffer.get(firstSubPacketPos)==STUFFING_BYTE) {
				firstSubPacketPos++;
			}
			int scanEnd=offset - 4 + Math.max(size, MINIMUM_VIDEO_PACKET_SIZE-9-headerLength);
			for (int i=offset; i<scanEnd; i++) {
				if (buffer.get(i)==BYTE0 && buffer.get(i+1)==BYTE0 && buffer.get(i+2)==BYTE1 && buffer.get(i+3)==SEQUENCE_PACKET) {
					return i;
				}
			}
		}
		return -1;
	}
	/*public static MpegVideoPacket findSequence(String fileName, ByteBuffer buffer) throws IOException {
		RandomAccessFile file=new RandomAccessFile(fileName,"rwd");
		file.readFully(buffer);
		file.close();
		int firstVideoPacket=-1;
		for (int i=0; i<buffer.length-3; i++) {
			if (buffer[i]==0 && buffer[i+1]==0 && buffer[i+2]==1 && buffer[i+3]==(byte)0xE0) {
				firstVideoPacket=i;
				break;
			}
		}
		if (firstVideoPacket >= 0) {
			MpegVideoPacket videoPacket = new MpegVideoPacket();
			videoPacket.parse(buffer, firstVideoPacket, buffer.length - firstVideoPacket);
			if (videoPacket.isVideoSequence()) {
				return videoPacket;
			}
		}
		return null;
	}*/

	public static String getAspectRatioText(int aspectRatioNumber) {
		if (aspectRatioNumber <= 4) {
			return ASPECT_RATIO_TEXT[aspectRatioNumber];
		} else {
			return "";
		}
	}
	public static int getTemporalReference(byte buffer[], int offset) {
		return ((buffer[offset+4] & 0xFF) << 2) + ((buffer[offset+5] & 0xC0) >>> 6);
	}
	public static boolean isPes(File file) {
		String fileName=file.getName().toLowerCase();
		return fileName.endsWith(".mpg") || fileName.endsWith(".m2p");
	}
	
	public void parse(ByteBuffer packet, int offset, int size) {
		if (size < MINIMUM_VIDEO_PACKET_SIZE) {
			return;
		}
		length=packet.getShort(offset + 4) & 0xffff;
		
		int ptsPosition=MpegUtil.getPtsPosition(packet, offset);
		if (ptsPosition > 0) {
			//PTS
			pts=MpegUtil.getPESTime(packet,offset + ptsPosition);
			byte ptsDtsFlag=packet.get(offset + 7);
			if ((ptsDtsFlag & DTS_PRESENT) == DTS_PRESENT) {
				//DTS
				dts=MpegUtil.getPESTime(packet,offset + 14);
			}
		} else {
			pts=0;
			dts=0;
		}
		headerLength=packet.get(offset+8) & 0xFF;
		
		//See if this packet starts with a video sequence header
		int firstSubPacketPos = offset + 9 + headerLength;
		while (packet.get(firstSubPacketPos)==STUFFING_BYTE) {
			firstSubPacketPos++;
		}
		sequencePacketStart=findSequence(packet, offset, size);
		if (sequencePacketStart !=-1) {
			sequenceHelper.parse(new MuxBuffer(packet), sequencePacketStart);
		} /*else {
			width=height=aspectRatioType=frameRateType=videoPacketStart=0;
		}*/
	}

	public long getDts() {
		return dts;
	}

	public int getHeight() {
		return sequenceHelper.getHeight();
	}

	public int getLength() {
		return length;
	}

	public long getPts() {
		return pts;
	}

	public int getWidth() {
		return sequenceHelper.getWidth();
	}

	public int getAspectRatioType() {
		return sequenceHelper.getAspectRatioType();
	}

	public int getFrameRateType() {
		return sequenceHelper.getFrameRateType();
	}

	public boolean isVideoSequence() {
		return sequencePacketStart !=-1;
	}
	
	/*public int getMuxRate() {
		if (this.height >= 720) {
			return 20000;
		} else {
			return 3303;
		}
	}*/
	public int getSequencePacketStart() {
		return this.sequencePacketStart;
	}
	public int getHeaderLength() {
		return headerLength;
	}
	public SequenceHelper getSequenceHelper() {
		return sequenceHelper;
	}
}
