package util;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.nio.ByteBuffer;
import java.nio.channels.FileChannel;
import java.util.HashMap;

import common.MuxBuffer;
import common.Util;
import common.mpeg.MpegUtil;
import common.mpeg.SequenceHelper;



/**
 * Dumps all the goodies from a PES stream
 * Very useful for debugging
 */
public class PESAnalizer {
	private static final byte STUFFING_BYTE=(byte)0xFF;
	private static final Integer EOF=new Integer(0xB9);
	//Largest packet a stream can have
	private ByteBuffer packet = ByteBuffer.allocateDirect(256*256 + 6); //+6 for my header + 6 more for next header
	//private byte packet2[]=new byte[packet.length];
	private static char hex[]={'0','1','2','3','4','5','6','7','8','9','A','B','C','D','E','F'};
	
	private static String bufferBoundTypes[]={"Audio","Video"};
	
	private HashMap packetDecoders=new HashMap();
	
	interface PacketDecoder {
		public String getPacketName();
		public int getPacketLength(ByteBuffer packet, int offset);
		public String getExtendInfo(ByteBuffer packet, int offset);
	}
	
	class PackHeaderPacket implements PacketDecoder {
		long lastScr;
		
		public String getPacketName() {
			return "Program Stream Pack";
		}
		public int getPacketLength(ByteBuffer packet, int offset) {
			int bytes=14+(packet.get(offset+13)&0x07);
			return bytes;
		}
		public String getExtendInfo(ByteBuffer packet, int offset) {
			StringBuffer out=new StringBuffer();
			out.append("SCR: ");
			long scr[]=MpegUtil.getSCRTime(packet, offset + 4);
			out.append(MpegUtil.pesTimeToString(scr[0]));
			out.append("|");
			out.append(scr[1]);
			out.append(" Mux Rate: ");
			out.append(MpegUtil.getMuxRate(packet, offset + 10));
			out.append(" SCR detlta: ");
			long scrDelta=scr[0]-lastScr;
			out.append(scrDelta);
			if (scrDelta > 2880) {
				System.out.println("Arg");
			}
			
			int stuffingBytes=(packet.get(offset+13)&0x07);
			if (stuffingBytes > 0) {
				out.append(" Stuffing: ");
				out.append(stuffingBytes);
			}
			lastScr=scr[0];
			return out.toString();
		}
	}
	
	class GenericPacket implements PacketDecoder {
		String packetName;
		
		public GenericPacket(String packetName) {
			this.packetName=packetName;
		}
		public String getPacketName() {
			return packetName;
		}
		public int getPacketLength(ByteBuffer packet, int offset) {
			return 6 + get2ByteInt(packet, offset+4);
		}
		public String getExtendInfo(ByteBuffer packet, int offset) {
			return null;
		}
	}
	class ExtendedPacket extends GenericPacket {
		long lastPts=0;
		PackHeaderPacket packetHeaderPacket;
		
		public ExtendedPacket(String packetName, PackHeaderPacket packetHeaderPacket) {
			super(packetName);
			this.packetHeaderPacket=packetHeaderPacket;
		}
		public String getExtendInfo(ByteBuffer packet, int offset) {
			byte ptsDtsFlag=packet.get(offset + 7);
			StringBuffer out=new StringBuffer();
			
			if ((ptsDtsFlag & MpegUtil.PTS_PRESENT) == MpegUtil.PTS_PRESENT) {
				//PTS
				long pts=lastPts=MpegUtil.getPESTime(packet,offset + 9);
				out.append("PTS: ");
				out.append(MpegUtil.pesTimeToString(pts));
				//out.append(" " + pts + " ");
				if ((ptsDtsFlag & MpegUtil.DTS_PRESENT) == MpegUtil.DTS_PRESENT) {
					//DTS
					long dts=MpegUtil.getPESTime(packet,offset + 14);
					out.append(" DTS: ");
					out.append(MpegUtil.pesTimeToString(dts));
				}
				if (packetHeaderPacket.lastScr > pts) {
					out.append(" LATE: ");
					out.append(packetHeaderPacket.lastScr-pts);
				}
			}
			return out.toString();
		}
		public long getLastPts() {
			return lastPts;
		}
	}
	
	class VideoPacket extends ExtendedPacket {
		private SequenceHelper sequenceHelper = new SequenceHelper();
		public VideoPacket(PackHeaderPacket packetHeaderPacket) {
			super("Video Packet", packetHeaderPacket);
		}
		public String getExtendInfo(ByteBuffer packet, int offset) {
			String extended=super.getExtendInfo(packet, offset);
			StringBuffer out;
			if (extended == null) {
				out=new StringBuffer();
			} else {
				out=new StringBuffer(extended);
				out.append(' ');
			}
			
			int headerLength=MpegUtil.getPesHeaderLength(packet, offset);
			
			//See if this packet starts with a video sequence header
			int firstSubPacketPos = offset + 9 + headerLength;
			while (packet.get(firstSubPacketPos)==STUFFING_BYTE) {
				firstSubPacketPos++;
			}
			MuxBuffer buffer=new MuxBuffer(packet);
			if (packet.get(firstSubPacketPos + 2)==01 && packet.get(firstSubPacketPos + 3)==(byte)0xB3) {
				sequenceHelper.parse(buffer, firstSubPacketPos);
				out.append(sequenceHelper.toString());

			}
			
			int length=getPacketLength(packet, offset);
			for (int i=headerLength; i<length -3; i++) {
				if (packet.get(i)==0 && packet.get(i+1)==0 && packet.get(i+2)==1) {
					if (packet.get(i+3)==0) {
						int temporalRef=MpegUtil.getTemporalReference(buffer, i);
						out.append(" Temp. Ref: ");
						out.append(temporalRef);
						int pictureType=(packet.get(i+5) & 0x38) >>> 3;
						switch (pictureType) {
							case 1:
								out.append(" I");
								break;
							case 2:
								out.append(" P");
								break;
							case 3:
								out.append(" B");
								break;
							default:
								out.append(" ?");
						}
						break;
					} else if (packet.get(i+3)==(byte)0xB8) {
						//Dump the GOP time
						out.append(' ');
						out.append((packet.get(i+4) & 0xFF)>>>3);
						out.append(':');
						int min=((packet.get(i+4) & 0x07) << 3) + ((packet.get(i+5) & 0xE0) >>> 5);
						out.append(min);
						int sec=((packet.get(i+5) & 0x0F) << 2) + ((packet.get(i+6) & 0xC0) >>> 6);
						out.append(':');
						out.append(sec);
						out.append('.');
						out.append((packet.get(i+6) & 0x3F));
					}
				}
			}
			
			if (out.length() != 0) {
				return out.toString();
			} {
				return null;
			}
			
		}
	}
	class PrivateStream1Packet extends ExtendedPacket {
		VideoPacket videoPacket;
		public PrivateStream1Packet(PackHeaderPacket packetHeaderPacket, VideoPacket videoPacket) {
			super("Private Stream 1", packetHeaderPacket);
			this.videoPacket=videoPacket;
		}
		public String getExtendInfo(ByteBuffer packet, int offset) {
			String extended=super.getExtendInfo(packet, offset);
			StringBuffer out;
			if (extended == null) {
				out=new StringBuffer();
			} else {
				out=new StringBuffer(extended);
			}
			int headerLength=MpegUtil.getPesHeaderLength(packet, offset);
			int lpcmpStart=offset + headerLength + 9;
			out.append(" SubStream: ");
			out.append(getHexNumber(packet.get(lpcmpStart) & 0xFF));
			out.append(" Frame Hdrs: ");
			out.append(getHexNumber(packet.get(lpcmpStart + 1) & 0xFF));
			out.append(" Frame Offset: ");
			out.append(getHexNumber(packet.get(lpcmpStart + 2) & 0xFF));
			out.append(getHexNumber(packet.get(lpcmpStart + 3) & 0xFF));
			out.append(" Offset: ");
			long delta=this.lastPts - videoPacket.lastPts;
			if (delta < 0) {
				out.append('-');
				delta *=-1;
			}
			out.append(MpegUtil.pesTimeToString(delta));
			/*out.append(" Frame #: ");
			out.append(getHexNumber(packet[lpcmpStart + 4] & 0x1F));
			out.append(" Channels: ");
			out.append(getHexNumber((packet[lpcmpStart + 5] & 0x07) + 1));
			out.append(" Range X,Y: ");
			out.append(getHexNumber((packet[lpcmpStart + 6] & 0xE0) >>> 5));
			out.append(',');
			out.append(getHexNumber(packet[lpcmpStart + 6] & 0x1F));*/
			//MpegUtil.printHex(packet, lpcmpStart, 16);
			if (out.length() != 0) {
				return out.toString();
			} {
				return null;
			}
		}
	}
	class PrivateStream2Packet extends ExtendedPacket {
		public PrivateStream2Packet(PackHeaderPacket packetHeaderPacket) {
			super("Private Stream 2", packetHeaderPacket);
		}
		public String getExtendInfo(ByteBuffer packet, int offset) {
			String extended=super.getExtendInfo(packet, offset);
			StringBuffer out;
			if (extended == null) {
				out=new StringBuffer();
			} else {
				out=new StringBuffer(extended);
			}
			for (int i=0;i<20;i++) {
				Util.appendHex(packet.get(offset+i), out);
				out.append(' ');
			}
			if (out.length() != 0) {
				return out.toString();
			} {
				return null;
			}
		}
	}
	class SystemHeaderPacket extends GenericPacket {
		public SystemHeaderPacket() {
			super("System Header");
		}
		public String getExtendInfo(ByteBuffer packet, int offset) {
			StringBuffer out=new StringBuffer();
			int rate=((packet.get(offset+6) & 0x7F) << 15) //15-22
				| ((packet.get(offset+7) & 0xFF) << 7)   //7-14
				| ((packet.get(offset+8) & 0xFF) >>> 1); //0-6
			out.append(" Rate: ");
			out.append(rate);
			out.append(" Audio Streams: ");
			out.append(packet.get(offset+9) >>> 2);
			out.append(" CPS Flag: ");
			out.append(packet.get(offset+9) & 1);
			out.append(" Audio Lock: ");
			out.append((0x80 & packet.get(offset+10)) >>> 7);
			out.append(" Video Lock: ");
			out.append((0x40 & packet.get(offset+10)) >>> 6);
			out.append(" Video Streams: ");
			out.append(0x1F & packet.get(offset+10));
			out.append(" Rate Restricted: ");
			out.append((0x80 & packet.get(offset+11)) >>> 7);
			
			out.append(" {");
			for (int i=12; (packet.get(i) & 0x80)==0x80;i=i+3) {
				out.append(getHexNumber(packet.get(i)&0xFF));
				//out.append(": ");
				int bufferBoundType=(packet.get(i+1) & 0x20) >>> 5;
				//out.append(bufferBoundTypes[bufferBoundType]);
				out.append(": ");
				int bufferSize=(((packet.get(i+1) & 0x1F) << 8) | (packet.get(i+2) & 0xFF)) / (bufferBoundType==0? 8:1);
				out.append(bufferSize);
				out.append("k,");
				//out.append(',');
			}
			out.append('}');
			if (out.length() != 0) {
				return out.toString();
			} {
				return null;
			}
		}
	}
	public PESAnalizer() {
		PackHeaderPacket packetHeaderPacket=new PackHeaderPacket();
		VideoPacket videoPacket=new VideoPacket(packetHeaderPacket);
		packetDecoders.put(new Integer(0xE0), videoPacket);
		packetDecoders.put(new Integer(0xC0), new ExtendedPacket("Audio Stream", packetHeaderPacket));
		packetDecoders.put(new Integer(0xBA), packetHeaderPacket);
		packetDecoders.put(new Integer(0xBB), new SystemHeaderPacket());//Always follows BA
		packetDecoders.put(new Integer(0xBD), new PrivateStream1Packet(packetHeaderPacket, videoPacket));
		packetDecoders.put(new Integer(0xBE), new GenericPacket("Padding Stream"));
		packetDecoders.put(new Integer(0xBF), new PrivateStream2Packet(packetHeaderPacket));
		packetDecoders.put(new Integer(0xFD), new ExtendedPacket("VC1/AVC Stream", packetHeaderPacket));
	}
	
	/**
	 * @param args
	 */
	public static void main(String[] args) {
		//String title="C:/Tivo/Desperate Housewives - Beautiful Girls.mpg";
		//String title="C:/Tivo/Desperate Housewives - Beautiful Girls (02).mpg";
		//String title="C:/Tivo/Tombstone.mpg";
		//String title="E:/World of Adventure Sports - Kayaking in Quebec.mpg";
		//String title="C:/Tivo/South Park - Cartoon Wars.mpg";
		//String title="C:/Tivo/Desperate Housewives - The Miracle Song.mpg";
		//String title="C:/Tivo/The Matrix Reloaded.mpg";
		//String title="C:/Tivo/The Matrix Reloaded (02).mpg";
		//String title="C:/Tivo/South Park.mpg";
		//new PESAnalizer().analize("C:/Tivo/South Park-Jakovasaurs.mpg");
		//new PESAnalizer().analize("M:/The Terminal.m2p");
		//new PESAnalizer().analize("C:/Tivo/Mash[pes].m2p");
		//new PESAnalizer().analize("C:/Tivo/Scary Movie.mpg");
		//new PESAnalizer().analize("C:/Tivo/South Park[mpeg-2].mpg");
		//new PESAnalizer().analize("C:/Tivo/Platoon-.mpg");
		//new PESAnalizer().analize("C:/Tivo/Platoon- (02).mpg");
		//new PESAnalizer().analize("C:/Tivo/Platoon-[mpeg-2].mpg");
		//String title="C:/Tivo/CSI_ Crime Scene Investigation - Monster in the Box.mpg";
		//String title="C:/Tivo/Desperate Housewives - The Little Things You Do Together.mpg";
		//String title="C:/Tivo/World of Adventure Sports.mpg";
		//String title="E:/World of Adventure Sports.mpg";
		//String title="E:/World.m2p";
		//String title="C:/Tivo/World.mpg";
		//String title="E:\\South Park - Tweek vs. Craig.mpg";
		//String title="C:\\Tivo\\South Park - Fat Butt and Pancake Head.mpg";
		//String title="C:/Tivo/finger.mpg";
		//String title="E:/Jump for Joy.mpg";
		//String title="E:/Burn Notice - Pilot.mpg";
		//String title="E:\\South Park - Timmy 2000.mpg";
		//String title="C:\\Documents and Settings\\Dustin Burckhardt\\workspace\\HdTranscode\\test_0.mpg";
		//String title="C:/Tivo/Test/Grey's Anatomy - Some Kind of Miracle.mpg";
		//String title="C:\\Tivo\\chunk.mpg";
		//String title="F:/chunk.mpg";
		//String title="i:/HVDVD_TS/FEATURE_1.EVO";
		//String title="H:/VIDEO_TS/VTS_06_1.VOB";
		//String title="C:/Walk the line_e0.mpg";
		//String title="C:/CSI_ Crime Scene Investigation - Monster in the Box.mpg";
		//String title="C:\\Test\\chunk.mpg";
		//String title="C:/Terminator.mpg";
		//String title="F:/test_fd55.vc1.mpg";
		//String title="H:/Top Gun.mpg";
		String title="C:/Lost.mpg";
		if (args.length==1) {
			title=args[0];
		}
		new PESAnalizer().analyze(title);
	}
	
	public void analyze(String fileName) {
		try {
			FileChannel in = new FileInputStream(fileName).getChannel();
			long position;
			int bytes=readFullPacket(in,0);
			position = bytes;
			while (bytes != -1) {
				
				Integer streamType=Integer.valueOf(packet.get(3)&0xff);
				if (streamType.equals(EOF)) {
					return;
				}
				
				PacketDecoder packetDecoder=(PacketDecoder)packetDecoders.get(streamType);
				if (packetDecoder==null) {
					packetDecoder=new GenericPacket("Unknown packet type: " + getHexNumber(streamType.intValue()));
					System.out.println(packetDecoder.getPacketName());
				}
				
				int packetSize=packetDecoder.getPacketLength(packet, 0);
				String extendedInfo = packetDecoder.getExtendInfo(packet, 0);
				System.out.println(packetDecoder.getPacketName() + ": " + packetSize + (extendedInfo==null?"":" : " + extendedInfo));
				
				if (!isHeader(packet, packetSize)) {
					System.out.println("Error in packet size attempting to fix... " + packetDecoder.getPacketName() +":" + packetSize);
					
					for (packetSize=6;packetSize<packet.capacity(); packetSize++) {
						if (isHeader(packet, packetSize) && packet.get(packetSize+3) > (byte)0xB0) {
							PacketDecoder testPacketDecoder=(PacketDecoder)packetDecoders.get(streamType);
							if (testPacketDecoder != null && isHeader(packet, packetSize + testPacketDecoder.getPacketLength(packet, packetSize))) {
								System.out.println("Found packet at: " + packetSize);
								break;
							}
							
						}
					}
				}
				
				int bytesRemaining=packet.capacity()-packetSize;
				packet.position(packetSize);
				packet.compact();
				//System.arraycopy(packet,packetSize, packet2, 0, bytesRemaining);
				/*byte temp[]=packet;
				packet=packet2;
				packet2=temp;*/
				bytes=readFullPacket(in, bytesRemaining);
				position += bytes;
			}
			in.close();
			
			
		} catch (FileNotFoundException fnfe) {
			System.out.println("File Not Found: " + fileName);
		} catch(IOException ioe) {
			System.out.println(ioe);
			ioe.printStackTrace();
		}
	}
	
	private int readFullPacket(FileChannel in, int bytes) throws IOException {
		bytes=in.read(packet);
		/*int read=read=in.read(packet, bytes, packet.capacity() - bytes);
		if (read > 0) {
			bytes += read;
		}
		if (bytes != packet.position() && read != -1) {
			readFullPacket(in, bytes);
		}*/
		return bytes;
	}
	
	private int get2ByteInt(ByteBuffer packet, int offset) {
		return  packet.getShort(offset) & 0xffff ;
	}
	
	private String getHexNumber(int number) {
		StringBuffer sb=new StringBuffer(8);
		sb.append(hex[(number&0xF0)>>4]);
		sb.append(hex[number&0xF]);
		return sb.toString();
	}
	private boolean isHeader(ByteBuffer packet, int offset) {
		return packet.get(offset)==0 && packet.get(offset+1)==0 && packet.get(offset+2) == 1;
	}
	
}

